---@type Tinkr
local Tinkr,
---@class Bastion
Bastion = ...

---@class Bastion.ObjectManager
---@field _lists table<string, {list: Bastion.List, cb:fun(Object: WowGameObject): any}>
---@field objects WowGameObject[]
---@field objectList table<string, WowGameObject>
local ObjectManager = {}
ObjectManager.__index = ObjectManager

---@param interval? number
function ObjectManager:New(interval)
    ---@class Bastion.ObjectManager
    local self = setmetatable({}, ObjectManager)
    self.interval = interval or 0.1
    self.nextUpdate = 0
    self._lists = {}
    self.objects = {}
    self.objectList = {}
    self.range = false
    self.lastStart = 0
    self.lastStop = 0
    self.lastDuration = 0
    self.refreshing = false


    self.activeEnemies = Bastion.List:New()
    self.afflicted = Bastion.List:New()
    self.critters = Bastion.List:New()
    self.enemies = Bastion.List:New()
    self.explosives = Bastion.List:New()
    self.friends = Bastion.List:New()
    self.incorporeal = Bastion.List:New()
    self.missiles = Bastion.List:New()
    self.others = Bastion.List:New()
    self.outOfRange = Bastion.List:New()

    return self
end

---@param range number
function ObjectManager:SetRange(range)
    self.range = type(range) == "number" and range or self.range
end

-- Register a custom list with a callback
---@param name string
---@param cb fun(object: TinkrObjectReference): boolean | any
---@return Bastion.List | false
function ObjectManager:RegisterList(name, cb)
    if self._lists[name] then
        return false
    end

    self._lists[name] = {
        list = Bastion.List:New(),
        cb = cb,
    }

    return self._lists[name].list
end

-- reset custom lists
---@return nil
function ObjectManager:ResetLists()
    for _, list in pairs(self._lists) do
        list.list:clear()
    end
end

function ObjectManager:Reset()
    self.objects = {}
    self.objectList = {}

    self.activeEnemies:clear()
    self.afflicted:clear()
    self.critters:clear()
    self.enemies:clear()
    self.explosives:clear()
    self.friends:clear()
    self.incorporeal:clear()
    self.missiles:clear()
    self.others:clear()
    self.outOfRange:clear()
    self:ResetLists()
end

-- Refresh custom lists
---@param object WowGameObject
---@return nil
function ObjectManager:EnumLists(object)
    for _, list in pairs(self._lists) do
        local r = list.cb(object)
        if r then
            list.list:push(r)
        end
    end
end

-- Get a list
---@param name string
---@return Bastion.List
function ObjectManager:GetList(name)
    return self._lists[name].list
end

function ObjectManager:IsRefreshing()
    return self.refreshing
end

function ObjectManager:InObjectList(objectGuid)
    return self.objectList[objectGuid] ~= nil
end

-- Refresh all lists
---@param force? boolean
function ObjectManager:Refresh(force)
    if self:IsRefreshing() or (self.nextUpdate > GetTime() and not force) or (not Bastion.Enabled and not force) then
        return
    end

    self.refreshing = true
    self.lastStart = GetTime()
    self.nextUpdate = self.lastStart + self.interval

    self:Reset()
    local objects = Objects()

    for _, object in pairs(objects) do
        local objectGuid = object:guid()
        local objectType = object:type()
        if objectGuid ~= false and objectType ~= false then
            self.objectList[objectGuid] = object

            table.insert(self.objects, object)
            self:EnumLists(object)
            if ({ [5] = true, [6] = true, [7] = true })[objectType] then
                if objectType == 5 and object:creatureType() == 8 then
                    self.critters:push(Bastion.Globals.UnitManager:GetObject(objectGuid) or
                        Bastion.Globals.UnitManager:SetObject(Bastion.Unit:New(object), objectGuid))
                elseif object:id() == 204560 --[[ unit:GetID() == 204560 ]] then
                    self.incorporeal:push(Bastion.Globals.UnitManager:GetObject(objectGuid) or
                        Bastion.Globals.UnitManager:SetObject(Bastion.Unit:New(object), objectGuid))
                elseif object:id() == 204773 then
                    self.afflicted:push(Bastion.Globals.UnitManager:GetObject(objectGuid) or
                        Bastion.Globals.UnitManager:SetObject(Bastion.Unit:New(object), objectGuid))
                elseif object:id() == 120651 then
                    self.explosives:push(Bastion.Globals.UnitManager:GetObject(objectGuid) or
                        Bastion.Globals.UnitManager:SetObject(Bastion.Unit:New(object), objectGuid))
                elseif UnitInParty(object:unit()) or UnitInRaid(object:unit()) or UnitIsUnit("player", object:unit()) then
                    self.friends:push(Bastion.Globals.UnitManager:GetObject(objectGuid) or
                        Bastion.Globals.UnitManager:SetObject(Bastion.Unit:New(object), objectGuid))
                elseif UnitCanAttack("player", object:unit()) then
                    local unit = Bastion.Globals.UnitManager:GetObject(objectGuid) or
                        Bastion.Globals.UnitManager:SetObject(Bastion.Unit:New(object), objectGuid)
                    self.enemies:push(unit)
                    if unit:IsAffectingCombat() or unit:InCombatOdds() > 80 then
                        self.activeEnemies:push(unit)
                    end
                else
                    self.others:push(object)
                    --[[ self.others:push(Bastion.UnitManager:GetObject(objectGuid) or
                        Bastion.UnitManager:SetObject(Bastion.Unit:New(object, objectGuid))) ]]
                end
            end
        end
    end

    --[[ local missiles = Missiles()
    if type(missiles) == "table" then
        for _, missile in pairs(missiles) do
            self.missiles:push(missile)
        end
    end ]]
    self.lastStop = GetTime()
    self.lastDuration = self.lastStop - self.lastStart
    self.refreshing = false
end

Bastion.ObjectManager = ObjectManager
