---@type Tinkr
local Tinkr,
---@class Bastion
Bastion = ...

---@class Bastion.UnitManager.CustomUnit
---@field unit Bastion.Cacheable | Bastion.Unit
---@field cb fun(): Bastion.Unit

-- Create a new UnitManager class
---@class Bastion.UnitManager
---@field units table<string, Bastion.Unit>
---@field customUnits table<string, Bastion.UnitManager.CustomUnit>
---@field objects table<string, Bastion.Unit>
---@field cache Bastion.Cache
local UnitManager = {
}

local tokenPrefixes = {
    "^arena",
    "^boss",
    "^focus",
    "^mouseover",
    "^none",
    "^party",
    "^partypet",
    "^pet",
    "^player",
    "^raid",
    "^raidpet",
    "^target",
    "^vehicle",
    "^nameplate",
    "^spectated",
    "^softenemy",
    "^softfriend",
    "^anyenemy",
    "^anyfriend",
    "^anyinteract",
    "^npc",
    "^questnpc",
}

---@param token string
local validateToken = function(token)
    for i = 1, #tokenPrefixes do
        if string.match(token, tokenPrefixes[i]) then
            return true
        end
    end
    return false
end

---@param k UnitId
function UnitManager:__index(k)
    if k == "none" then
        return self.objects.none
    end

    if UnitManager[k] ~= nil then
        return UnitManager[k]
    end

    if rawget(self, k) ~= nil then
        return rawget(self, k)
    end

    local key = k or "none"

    -- if custom unit exists return it or a new one if the cache has expired.
    if self.customUnits[key] then
        if not self.cache:IsCached(key) then
            self.customUnits[key].unit:Update()
            self.cache:Set(key, self.customUnits[key].unit, 0.5)
        end

        return self.customUnits[key].unit
    end

    local kguid = type(key) == "string" and validateToken(key) and ObjectGUID(key) or key

    if kguid ~= nil and self.objects[kguid] then
        return self.objects[kguid]
    end
    if kguid ~= nil and self.objects[kguid] == nil then
        local obj = Object(kguid)
        if obj then
            self.objects[kguid] = Bastion.Unit:New(obj, kguid)
            return self:SetObject(Bastion.Unit:New(obj, kguid))
        end
    end

    --if obj and objGuid then
    --if self.objects[objGuid] == nil then
    --self:SetObject(Bastion.Unit:New(obj, objGuid))
    --end
    --[[ return self:Get(objGuid) or self:SetObject(Bastion.Unit:New(obj), objGuid) ]]
    --[[ if self.objects[obj:guid()] ~= nil then
            return self.objects[obj:guid()]
        else
            return self:SetObject(Bastion.Unit:New(obj), obj:guid() or nil)
        end ]]
    --end

    return self.objects.none
end

-- Constructor
---@return Bastion.UnitManager
function UnitManager:New()
    ---@class Bastion.UnitManager
    local self = setmetatable({}, UnitManager)
    self.units = {}
    self.customUnits = {}
    self.objects = {
        none = Bastion.Unit:New(),
    }
    self.cache = Bastion.Cache:New()
    return self
end

-- Get or create a unit
---@param token TinkrObjectReference
function UnitManager:Get(token)
    --[[ local unitObject = Object(token)
    local tguid = unitObject and unitObject:guid() or "none"
    if self.objects[tguid] == nil then
        if tguid == 'none' then
            self.objects.none = Bastion.Unit:New()
        else
            self.objects[tguid] = Bastion.Unit:New(unitObject)
        end
    end

    return Bastion.Refreshable:New(self.objects[tguid], function()
        local tguid = ObjectGUID(token) or "none"
        if self.objects[tguid] == nil then
            if token == 'none' then
                self.objects.none = Bastion.Unit:New()
            else
                self.objects[tguid] = Bastion.Unit:New(Object(tguid))
            end
        end
        return self.objects[tguid]
    end) ]]

    local tguid = ObjectGUID(token)

    if tguid and self.objects[tguid] == nil then
        local obj = Object(tguid)
        if obj then
            self.objects[tguid] = Bastion.Unit:New(obj, tguid)
        end
    end

    return Bastion.Refreshable:New(self.objects[tguid], function()
        local tguid = ObjectGUID(token) or "none"
        if self.objects[tguid] == nil then
            local obj = Object(tguid)
            if obj then
                self.objects[tguid] = Bastion.Unit:New(obj, tguid)
            end
        end
        return self.objects[tguid]
    end)
end

-- Get a unit by guid
---@param guid string | WowGameObject
---@return Bastion.Unit?
function UnitManager:GetObject(guid)
    return self.objects[guid]
end

-- Get a unit by guid or create a new one if it does not exist.
---@param unit string | WowGameObject
---@return Bastion.Unit?
function UnitManager:GetOrCreateObject(unit)
    local guid = type(unit) == "string" and unit or ObjectGUID(unit)
    if guid then
        local unitObj = self:GetObject(guid)
        if unitObj and not unitObj:IsValid() then
            local object = type(unit) ~= "string" and unit or Object(guid)
            if object then
                unitObj = self:SetObject(Bastion.Unit:New(object, guid))
            end
        end
        return unitObj
    end
end

-- Set a unit by guid
---@param unit Bastion.Unit
---@param guid? string
function UnitManager:SetObject(unit, guid)
    local tguid = guid or unit:GetGUID()
    if tguid then
        self.objects[tguid] = unit
        return self.objects[tguid]
    end
end

---@param unitGUID string
function UnitManager:ClearObject(unitGUID)
    self.objects[unitGUID] = nil
end

-- Create a custom unit and cache it for .5 seconds
---@generic V :Bastion.Unit
---@param token string
---@param cb fun(): Bastion.Unit
function UnitManager:CreateCustomUnit(token, cb)
    local unit = cb()
    local cachedUnit = Bastion.Cacheable:New(unit, cb)
    if unit == nil then
        error("UnitManager:CreateCustomUnit - Invalid unit: " .. token)
    end
    if self.customUnits[token] == nil then
        self.customUnits[token] = {
            unit = cachedUnit,
            cb = cb,
        }
    end

    self.cache:Set(token, cachedUnit, 0.5)
    return cachedUnit
end

function UnitManager:ClearInvalid()
    for k, object in pairs(self.objects) do
        if k ~= "none" and not self.objects[k]:IsValid() then
            self.objects[k] = nil
        end
    end
end

function UnitManager:ResetObjects()
    for k, object in pairs(self.objects) do
        if k ~= "none" and not Bastion.ObjectManager.objectList[k] then
            self.objects[k] = nil
        end
    end
end

---@description Enumerates all friendly units in the battlefield
---@param cb fun(unit: Bastion.Unit):boolean
---@return nil
function UnitManager:EnumFriends(cb)
    Bastion.ObjectManager.friends:each(function(unit)
        if cb(unit) then
            return true
        end
        return false
    end)
end

-- Enum Enemies (object manager)
---@param cb fun(unit: Bastion.Unit):boolean
---@return nil
function UnitManager:EnumEnemies(cb)
    Bastion.ObjectManager.activeEnemies:each(function(unit)
        if cb(unit) then
            return true
        end
        return false
    end)
end

-- Enum Units (object manager)
---@param cb fun(unit: Bastion.Unit):boolean
---@return nil
function UnitManager:EnumUnits(cb)
    Bastion.ObjectManager.enemies:each(function(unit)
        if cb(unit) then
            return true
        end
        return false
    end)
end

-- Enum Incorporeal (object manager)
---@param cb fun(unit: Bastion.Unit):boolean
---@return nil
function UnitManager:EnumIncorporeal(cb)
    Bastion.ObjectManager.incorporeal:each(function(unit)
        if cb(unit) then
            return true
        end
        return false
    end)
end

-- Get the number of enemies with a debuff
---@param spell Bastion.Spell
---@param range? number
---@return number
function UnitManager:GetNumEnemiesWithDebuff(spell, range)
    local count = 0
    if range == nil then
        range = spell:GetRange()
    end
    self:EnumEnemies(function(unit)
        if unit:GetAuras():FindMy(spell):IsUp() then
            count = count + 1
        end
        return false
    end)
    return count
end

-- Get the number of friends with a buff (party/raid members)
---@param spell Bastion.Spell
---@return number
function UnitManager:GetNumFriendsWithBuff(spell)
    local count = 0
    self:EnumFriends(function(unit)
        if unit:GetAuras():FindMy(spell):IsUp() then
            count = count + 1
        end
        return false
    end)
    return count
end

-- Get the number of friends alive (party/raid members)
---@return number
function UnitManager:GetNumFriendsAlive()
    local count = 0
    self:EnumFriends(function(unit)
        if unit:IsAlive() then
            count = count + 1
        end
        return false
    end)
    return count
end

-- Get the friend with the most friends within a given radius (party/raid members)
---@param radius number
---@return Bastion.Unit | nil
---@return Bastion.Unit[]
function UnitManager:GetFriendWithMostFriends(radius)
    local unit = nil
    local count = 0
    local friends = {}
    self:EnumFriends(function(u)
        if u:IsAlive() then
            local c = 0
            self:EnumFriends(function(other)
                if other:IsAlive() and u:GetDistance(other) <= radius then
                    c = c + 1
                end
                return false
            end)
            if c > count then
                unit = u
                count = c
                friends = {}
                self:EnumFriends(function(other)
                    if other:IsAlive() and u:GetDistance(other) <= radius then
                        table.insert(friends, other)
                    end
                    return false
                end)
            end
        end
        return false
    end)
    return unit, friends
end

-- Get the enemy with the most enemies within a given radius
---@return Bastion.Unit|nil, Bastion.Unit[]
function UnitManager:GetEnemiesWithMostEnemies(radius)
    local unit = nil
    local count = 0
    local enemies = {}
    self:EnumEnemies(function(u)
        if u:IsAlive() then
            local c = 0
            self:EnumEnemies(function(other)
                if other:IsAlive() and u:GetDistance(other) <= radius then
                    c = c + 1
                end
                return false
            end)
            if c > count then
                unit = u
                count = c
                enemies = {}
                self:EnumEnemies(function(other)
                    if other:IsAlive() and u:GetDistance(other) <= radius then
                        table.insert(enemies, other)
                    end
                    return false
                end)
            end
        end
        return false
    end)
    return unit, enemies
end

-- Find the centroid of the most dense area of friends (party/raid members) of a given radius within a given range
---@param radius number
---@param range number
---@return Bastion.Vector3 | nil
function UnitManager:FindFriendsCentroid(radius, range)
    local unit, friends = self:GetFriendWithMostFriends(radius)
    if unit == nil then
        return nil
    end

    local centroid = Bastion.Vector3:New(0, 0, 0)
    local zstart = -math.huge
    for i = 1, #friends do
        local p = friends[i]:GetPosition()
        centroid = centroid + p
        zstart = p.z > zstart and p.z or zstart
    end

    centroid = centroid / #friends

    if unit:GetPosition():Distance(centroid) > range then
        return unit:GetPosition()
    end

    local _, _, z = TraceLine(centroid.x, centroid.y, centroid.z + 5, centroid.x, centroid.y, centroid.z - 5, 0x100151)

    centroid.z = z + 0.01

    return centroid
end

-- Find the centroid of the most dense area of enemies of a given radius within a given range
---@param radius number
---@param range number
---@return Bastion.Vector3 | nil
function UnitManager:FindEnemiesCentroid(radius, range)
    local unit, enemies = self:GetEnemiesWithMostEnemies(radius)
    if unit == nil then
        return nil
    end

    local centroid = Bastion.Vector3:New(0, 0, 0)
    local zstart = -math.huge
    for i = 1, #enemies do
        local p = enemies[i]:GetPosition()
        centroid = centroid + p
        zstart = p.z > zstart and p.z or zstart
    end

    centroid = centroid / #enemies

    if unit:GetPosition():Distance(centroid) > range then
        return unit:GetPosition()
    end

    local _, _, z = TraceLine(centroid.x, centroid.y, centroid.z + 5, centroid.x, centroid.y, centroid.z - 5, 0x100151)

    centroid.z = z + 0.01

    return centroid
end

Bastion.UnitManager = UnitManager
